
object CPoint {
  
  def cartesian(x:Double,y:Double,c:Color):CPoint =
    new CartesianCPoint(x,y,c)
  
  def polar(r:Double,theta:Double,c:Color):CPoint = 
    if (r<0) 
      throw new Error("r negative") 
    else 
      new PolarCPoint(r,theta,c)




  private class CartesianCPoint (xpos:Double, ypos:Double, c:Color) extends CPoint {

    // delegate -- takes care of point-related operations
    val del:Point = Point.cartesian(xpos,ypos)

    // these methods can all be delegated
    
    def xCoord ():Double = del.xCoord()
    def yCoord ():Double = del.yCoord()
    def distanceFromOrigin ():Double = del.distanceFromOrigin()
    def angleWithXAxis ():Double = del.angleWithXAxis()
    def distance (q:CPoint):Double = del.distance(q)
    def isOrigin ():Boolean = del.isOrigin()

    def distance (q:Point):Double = del.distance(q)
    def add (q:Point):Point = del.add(q)


    // these method cannot be easily delegated
    // (they create new CPoints, or they rely on colors)

    def move (dx:Double,dy:Double):CPoint = 
      new CartesianCPoint(xpos+dx, ypos+dy,c)
    
    def add (q:CPoint):CPoint = 
      new CartesianCPoint(xpos+q.xCoord(),ypos+q.yCoord(),q.color())

    def rotate (t:Double):CPoint = 
      new CartesianCPoint(xpos*math.cos(t)-ypos*math.sin(t),
 			  xpos*math.sin(t)+ypos*math.cos(t),
			  c)
    
    def isEqual (q:CPoint):Boolean = 
      (xpos == q.xCoord()) && (ypos == q.yCoord()) && (c==q.color())

    def color ():Color = c

    def updateColor (nc:Color):CPoint = 
      new CartesianCPoint(xpos,ypos,nc)


    // BRIDGE METHODS

    def isEqual (q:Point):Boolean = q match {
      case cq:CPoint => isEqual(cq)
      case _ => false
    }


    // CANONICAL METHODS

    override def toString ():String = 
      "cartesian(" + xpos + "," + ypos + "," + c + ")"

    override def equals (other : Any):Boolean = 
      other match {
	case that : CPoint => this.isEqual(that)
	case _ => false
      }

    override def hashCode ():Int = 
      41 * (
        41 * (
  	  41 + xpos.hashCode()
        ) + ypos.hashCode()
      ) + c.hashCode()
  }



  private class PolarCPoint (r:Double, theta:Double, c:Color) extends CPoint {

    // delegate
    val del:Point = Point.polar(r,theta)

    def xCoord ():Double = del.xCoord()
    def yCoord ():Double = del.yCoord()
    def distanceFromOrigin ():Double = del.distanceFromOrigin()
    def angleWithXAxis ():Double = del.angleWithXAxis()
    def distance (q:CPoint):Double = del.distance(q)
    def isOrigin ():Boolean = del.isOrigin()

    def distance (q:Point):Double = del.distance(q)
    def add (q:Point):Point = del.add(q)


    def move (dx:Double,dy:Double):CPoint = 
      new CartesianCPoint(xCoord()+dx, yCoord()+dy,c)
    
    def add (q:CPoint):CPoint = 
      new CartesianCPoint(xCoord()+q.xCoord(),yCoord()+q.yCoord(),q.color())

    def rotate (angle:Double):CPoint = 
      new PolarCPoint(r, theta+angle,c)

    private def normalize (angle:Double):Double = 
      if (angle >= 2*math.Pi)
	normalize(angle-2*math.Pi)
      else if (angle < 0)
	normalize(angle+2*math.Pi)
      else
	angle

    def isEqual (q:CPoint):Boolean = {
      (r == q.distanceFromOrigin()) && 
      (normalize(theta) == normalize(q.angleWithXAxis())) &&
      c==q.color()
    }

    def color ():Color = c

    def updateColor (nc:Color):CPoint = 
      new PolarCPoint(r,theta,nc)
      
    // BRIDGE METHODS

    def isEqual (q:Point):Boolean = q match {
      case cq:CPoint => isEqual(cq)
      case _ => false
    }

    // CANONICAL METHODS
    
    override def toString ():String = 
      "polar(" + r + "," + theta + "," + c + ")"
    
    override def equals (other : Any):Boolean = 
      other match {
	case that : CPoint => this.isEqual(that)
	case _ => false
      }

    override def hashCode ():Int = 
      41 * (
        41 * (
  	  41 + r.hashCode()
        ) + theta.hashCode()
      ) + c.hashCode()
  }
}



abstract class CPoint extends Point {

  def xCoord ():Double
  def yCoord ():Double
  def angleWithXAxis ():Double
  def distanceFromOrigin ():Double
  def distance (q:CPoint):Double
  def move (dx:Double,dy:Double):CPoint
  def add (q:CPoint):CPoint
  def rotate (theta:Double):CPoint
  def isEqual (q:CPoint):Boolean
  def isOrigin ():Boolean

  def color ():Color
  def updateColor (nc:Color):CPoint

  // bridge methods
  def distance (q:Point):Double
  def add (q:Point):Point
  def isEqual (q:Point):Boolean
}
